from os.path import join
from tempfile import TemporaryDirectory, gettempdir
from unittest import TestCase

from topiq import Message, MessageType, Subscriber, Topic


class Worker:

    def __init__(self, message: Message):
        self.message = message

    async def run(self):
        id = self.message.as_int()
        ...


class SubsriberTest(TestCase):

    def setUp(self):
        self.temporary_folder = TemporaryDirectory()
        message = Message(MessageType.BYTES, b'\x01')
        self.topic_file_path = join(self.temporary_folder.name, 'topic1.et')
        with open(self.topic_file_path, 'wb') as writer:
            writer.write(message.to_bytes())

    def tearDown(self):
        self.temporary_folder.cleanup()

    async def test_subscriber(self):
        subscriber = Subscriber(Topic('name'), Worker)
        await subscriber.send_int_message(1)
        self.assertIsNotNone(subscriber)



from asyncio import new_event_loop, sleep
from functools import reduce
from operator import iconcat
from re import split
from threading import current_thread
from typing import Iterable, Type
from lxg import Loggable

from ._topic_reader import _TopicReader
# für Volksaufklärung
from .message import Message

_partitions_global_counter = 0


class Partition(Loggable):

    def __init__(self, worker_class: Type, _topic_reader: _TopicReader, partition_id: int = None):
        super().__init__()
        global _partitions_global_counter
        _partitions_global_counter += 1
        self.partition_id = partition_id if partition_id else _partitions_global_counter
        self._topic_reader = _topic_reader
        self.worker_class = worker_class

    async def _read_topic(self):
        while True:
            try:
                self.info(f'Topic reader {self._topic_reader}')
                with self._topic_reader as topic_reader:
                    message = Message.parse(await topic_reader.read())
                    await self.worker_class(message).run()
            except BaseException as e:
                self.error(e)
                await sleep(1)

    def run(self):
        loop = new_event_loop()

        try:
            self.info(f'Start Embedded Topic Partition [{self.partition_id}]')
            current_thread().setName(
                f'Embedded Topic Partition [{self.partition_id}]')
            loop.run_until_complete(self._read_topic())
        except InterruptedError:
            pass
        finally:
            loop.close()



from os.path import join
from tempfile import TemporaryDirectory, gettempdir
from unittest import TestCase

from topiq._topic_reader import _TopicReader
from topiq._topic_writer import _TopicWriter
from topiq.message import Message
from topiq.message_type import MessageType


class TopicTest(TestCase):

    def setUp(self):
        self.temporary_folder = TemporaryDirectory()
        message = Message(MessageType.BYTES, b'\x01')
        self.topic_file_path = join(self.temporary_folder.name, 'topic1.et')
        with open(self.topic_file_path, 'wb') as writer:
            writer.write(message.to_bytes())

    def tearDown(self):
        self.temporary_folder.cleanup()

    def test_reader(self):
        _topic_reader = _TopicReader(topic_file_path=self.topic_file_path)
        self.assertIsNotNone(_topic_reader)

    def test_writer(self):
        with open(self.topic_file_path, 'wb') as writer:
            writer.truncate()
        _topic_writer = _TopicWriter(topic_file_path=self.topic_file_path)
        self.assertIsNotNone(_topic_writer)



from io import BufferedIOBase
from mmap import ACCESS_READ, ACCESS_WRITE, mmap
from os import stat
from os.path import exists
from typing import Optional

from lxg import Loggable

from ._topic_event import _TopicEvent
from ._topic_utils import _get_stop_offset_and_truncate


class _TopicReader(Loggable):

    def __init__(self, topic_file_path: str, offset_event: _TopicEvent = None, offset: int = 0):
        self.topic_file_path: str = topic_file_path
        self._stream: BufferedIOBase = None
        self._mmapio: mmap = None
        self._offset = offset
        self._offset_event = offset_event
        self._stop_offset = None

    def __enter__(self):
        try:
            self._stop_offset = _get_stop_offset_and_truncate(
                self.topic_file_path)
            self._stream = open(self.topic_file_path, 'r+')
            self._mmapio = mmap(self._stream.fileno(),
                                length=0, access=ACCESS_READ)
            self._mmapio.seek(self._offset)
            return self
        except BaseException as e:
            self.error(e)
            self.info(f'Topic file path {self.topic_file_path}')
            raise e

    def __exit__(self, *args, **kwargs):
        try:
            self._mmapio.close()
            self._stream.close()
        except BaseException as e:
            self.error(e)
            raise e
        finally:
            if len(args) > 0 and isinstance(args[0], BaseException):
                raise args[0].with_traceback(args[1])
        return self

    async def change_offset_event(self, offset: int):
        self._offset = offset
        self._offset_event.set()

    async def read_the_message(self, offset: int):
        self.info('read the message')
        
        # READ 8 BYTES
        self._mmapio.seek(offset)
        buffer = self._mmapio.read(8)
        payload_length = int.from_bytes(buffer, 'big')
        if payload_length > 8:
            return buffer + self._mmapio.read(payload_length - 8)
        return None

    async def read(self) -> Optional[bytes]:

        if self._stop_offset and self._offset < self._stop_offset:
            message = await self.read_the_message(self._offset)
            self._offset += len(message)
            return message

        await self._offset_event.async_wait()
        self._offset_event.clear()
        
        message = await self.read_the_message(self._offset)
        self._offset += len(message)
        return message



# Embedded topics

# Each Embedded topic can have multiple partitions; by using more partitions,
# the consumers of the messages (and the throughput) may be scaled and 
# concurrency of processing increased.

# On top of publish-subscribe with partitions, a point-to-point messaging 
# system is built, by putting a significant amount of logic into the 
# consumers (in the other messaging systems we’ve looked at, it was the 
# server that contained most of the message-consumed-by-one-consumer logic;
# here it’s the consumer).

# Each consumer in a consumer group reads messages from a number of dedicated
# partitions; hence it doesn’t make sense to have more consumer threads than 
# partitions. Messages aren’t acknowledged on server (which is a very 
# important design difference!), but instead message offsets processed by 
# consumers are written to Zookeeper, either automatically in the background,
# or manually.

from .version import get_version

VERSION = (0, 1, 1, 'beta', 1)

__version__ = get_version(VERSION)

from .consumer import *
from .message_type import *
from .message import *
from .partition import *
from .subscriber import *
from .topic_error import *
from .topic import *



from io import BufferedIOBase
from mmap import ACCESS_READ, ACCESS_WRITE, mmap
from os import chmod, stat
from os.path import exists

DEFAULT_NEXT_BLOCK_SIZE = 65535


def _truncate_to(file_path: str, new_file_size: int):
    with open(file_path, 'a') as writer:
        writer.truncate(new_file_size)


def _get_stop_offset_and_truncate(file_path):
    file_size = stat(file_path).st_size if exists(file_path) else 0
    if file_size < DEFAULT_NEXT_BLOCK_SIZE:
        _truncate_to(file_path, DEFAULT_NEXT_BLOCK_SIZE)
    with open(file_path, 'r+') as reader:
        offset = 0
        buffer = reader.buffer.read(8)
        payload_length = int.from_bytes(buffer, 'big')
        if (payload_length < 8):
            return 0
        while payload_length > 8 and payload_length < file_size:
            offset += payload_length
            reader.seek(payload_length - 8)
            buffer = reader.buffer.read(8)
            payload_length = int.from_bytes(buffer, 'big')
        return offset



from asyncio import get_event_loop, new_event_loop
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Event
from os.path import join
from threading import Thread
from time import sleep
from typing import Type

from lxg import Loggable

from ._topic_event import _TopicEvent
from ._topic_reader import _TopicReader
from ._topic_writer import _TopicWriter
from .message import Message
from .partition import Partition


class Topic(Loggable):

    def __init__(self, topic_name: str, worker_class: Type, concurrency: int = 1):
        self.worker_class = worker_class
        self.topic_name = topic_name
        self.concurrency = concurrency
        self._topic_reader = None
        self._topic_writer = None
        self.thread = None

    async def initialize(self, topic_data_folder: str):
        self._topic_reader = _TopicReader(
            join(topic_data_folder, f'{self.topic_name}.et'),
            _TopicEvent()
        )
        self._topic_writer = _TopicWriter(
            join(topic_data_folder, f'{self.topic_name}.et'))

        # Only one partition for concurrency level 1
        if self.thread is None:
            thread = Thread(
                target=self.run, args=())
            thread.daemon = True
            thread.start()

    async def send(self, message: Message):
        self.debug('send message')
        offset = self._topic_writer.write(message.to_bytes())
        await self._topic_reader.change_offset_event(offset)

    def run(self):
        executor = ThreadPoolExecutor(
            max_workers=self.concurrency,
        )

        for i in range(0, self.concurrency):
            executor.submit(Partition(self.worker_class,
                            self._topic_reader, i + 1).run)



def get_version(version=None):
    "Returns a PEP 386-compliant version number from VERSION."
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    main = get_main_version(version)
    sub = ''
    if version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version[3]] + str(version[4])

    return str(main + sub)


def get_main_version(version=None):
    "Returns main version (X.Y[.Z]) from VERSION."
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    return '.'.join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    """Returns a tuple of the django version. If version argument is non-empty,
    then checks for correctness of the tuple provided.
    """
    if version is None:
        from topiq import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    return version


def get_docs_version(version=None):
    version = get_complete_version(version)
    if version[3] != 'final':
        return 'dev'
    else:
        return '%d.%d' % version[:2]



from .message import Message
from .message_type import MessageType
from .topic import Topic


class Subscriber:

    def __init__(self, topic: Topic):
        self.topic = topic

    async def send_int_message(self, value: int):
        await self.topic.send(Message.from_int(value))

    def close(self):
        ...



from io import BufferedIOBase
from mmap import ACCESS_READ, ACCESS_WRITE, mmap
from os import chmod, stat
from os.path import exists
from lxg import Loggable

from ._topic_utils import _get_stop_offset_and_truncate

DEFAULT_TOPIC_FILE_LIMIT = 1073741824
DEFAULT_NEXT_BLOCK_SIZE = 65535


class _TopicWriter(Loggable):

    def __init__(self, topic_file_path: str, topic_file_size_limit: int = DEFAULT_TOPIC_FILE_LIMIT):
        self.topic_file_path: str = topic_file_path
        self.topic_file_size_limit: int = topic_file_size_limit
        self._stream: BufferedIOBase = None
        self._mmapio: mmap = None
        self._map_size = None
        self._offset = 0
        self._initial_size = 0  # Update on enter phase from file_size variable

    def _resize(self, new_map_size: int):
        if self._mmapio:
            self._mmapio.flush()
            self._mmapio.close()
        self._stream.truncate(new_map_size)
        self._mmapio = mmap(self._stream.fileno(),
                            length=0, access=ACCESS_WRITE)
        self._mmapio.seek(self._offset)

    def _init(self):
        self._offset = _get_stop_offset_and_truncate(self.topic_file_path)
        self._stream = open(self.topic_file_path, 'w+')
        chmod(self.topic_file_path, 0o777)
        self._map_size = (1 + self._offset //
                          DEFAULT_NEXT_BLOCK_SIZE) * DEFAULT_NEXT_BLOCK_SIZE
        self._resize(self._map_size)

    def __enter__(self):
        self._init()
        return self

    def __exit__(self, *args, **kwargs):
        try:
            self._mmapio.flush()
            self._mmapio.close()
            # self._stream.flush()
            # File size tune
            if (self._map_size - self._offset) < DEFAULT_NEXT_BLOCK_SIZE or \
                    self._offset == self._initial_size:
                self._stream.truncate(self._offset)
            self._stream.close()
        finally:
            if len(args) > 0 and isinstance(args[0], BaseException):
                raise args[0].with_traceback(args[1])
        return self

    def write(self, message: bytes) -> int:
        if not message:
            return
        if self._map_size is None:
            self._init()
        intented_length = self._offset + len(message)
        if intented_length > self.topic_file_size_limit:
            # Reset write position to the begin of the file
            self._mmapio.seek(0)
            self._offset = 0
        elif intented_length > self._map_size:
            self._map_size = (1 + intented_length //
                              DEFAULT_NEXT_BLOCK_SIZE) * DEFAULT_NEXT_BLOCK_SIZE
            self._resize(self._map_size)

        self._mmapio.write(message)

        result = self._offset
        self._offset += len(message)
        return result



from setuptools import setup, find_packages, Extension

with open('README.rst') as readme:
    __doc__ = readme.read()


# Provided as an attribute, so you can append to these instead
# of replicating them:
standard_exclude = ('*.py', '*.pyc', '*$py.class', '*~', '.*', '*.bak', '*.orig')
standard_exclude_directories = ('.*', 'CVS', '_darcs', './build', './dist', 'EGG-INFO', '*.egg-info')


# Dynamically calculate the version based on topiq.VERSION.
version = __import__('topiq').get_version()

setup(
    name='topiq',
    version=version,
    description='Lightweigt embedded topiqs (based on kafka ideas)',
    long_description=__doc__,
    license='Apache 2.0',
    author='Vanya Usalko',
    author_email='iusalko@eu.spb.ru',
    url='https://github.com/usalko/topiq',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
    ],
    zip_safe=False,
    install_requires=[],
)



from enum import Enum
from typing import Any


class MessageType(Enum):
    BYTES = 65
    INT = 73
    STR = 83
    FLOAT = 70
    BOOL = 66

    @staticmethod
    def to_bytes(message_type) -> bytes:
        return int(message_type.value).to_bytes(1, 'big')

    @staticmethod
    def from_bytes(message_type: bytes) -> Any:
        value = int.from_bytes(message_type, 'big')
        if MessageType.BYTES.value == value:
            return MessageType.BYTES
        if MessageType.INT.value == value:
            return MessageType.INT
        if MessageType.STR.value == value:
            return MessageType.STR
        if MessageType.FLOAT.value == value:
            return MessageType.FLOAT
        if MessageType.BOOL.value == value:
            return MessageType.BOOL
        return None



from unittest import TestCase

from topiq.message import Message
from topiq.message_type import MessageType


class MessageTest(TestCase):

    def test_simple(self):
        message = Message(MessageType.INT, -1)
        self.assertIsNotNone(message)

    def test_parse(self):
        message = Message(MessageType.BYTES, b'\x01')
        parsed_message = Message.parse(message.to_bytes())
        self.assertEquals(message.message_type, parsed_message.message_type)



from struct import pack, unpack
from typing import Any

from .topic_error import TopicError
from .message_type import MessageType


class Message:

    def __init__(self, message_type: MessageType = None, payload: bytes = None):
        self.message_type: MessageType = message_type
        self.payload: bytes = payload

    def to_bytes(self) -> bytes:
        entry = MessageType.to_bytes(self.message_type) + self.payload
        message_length = len(entry) + 16
        return message_length.to_bytes(8, 'big') + entry + message_length.to_bytes(8, 'little')

    @staticmethod
    def parse(message: bytes) -> Any:
        if len(message) < 16:
            raise TopicError('Not valid message, length less then 16')
        message_length_option_one = int.from_bytes(message[:8], 'big')
        message_length_option_two = int.from_bytes(message[-8:], 'little')
        if message_length_option_one != message_length_option_two:
            raise TopicError(
                'Not valid message, lengths are different (big-endian from the start and little-endian from the end)')
        if len(message) == 16:
            return Message()
        message_type = MessageType.from_bytes(message[8:9])
        if not(message_type is None):
            return Message(message_type, message[9:len(message) - 8])
        raise TopicError('Not valid message, unknown message type')

    @staticmethod
    def from_bytes(payload: bytes) -> Any:
        return Message(MessageType.BYTES, payload)

    @staticmethod
    def from_int(payload: int) -> Any:
        return Message(MessageType.INT, int(payload).to_bytes(8, 'big', signed=True))

    @staticmethod
    def from_str(payload: str) -> Any:
        return Message(MessageType.STR, payload.encode('utf-8'))

    @staticmethod
    def from_float(payload: float) -> Any:
        return Message(MessageType.FLOAT, bytearray(pack('f', payload)))

    @staticmethod
    def from_bool(payload: bool) -> Any:
        return Message(MessageType.BOOL, bytearray(pack('f', payload)))

    def as_int(self) -> int:
        if self.message_type in { MessageType.INT, MessageType.FLOAT }:
            if self.message_type == MessageType.INT:
                return int.from_bytes(self.payload, 'big', signed=True)
            if self.message_type == MessageType.FLOAT:
                return int(unpack('f', self.payload))
        raise TopicError(f'Sorry, but I can\'t read not number message. Message type is {self.message_type}')
